// Minimization of binary energies with graph cuts, using Veksler, Boykov & Kolmogorov's standard code

#ifndef MINIMIZATION_OF_MRF
#define MINIMIZATION_OF_MRF


// Old : MRF1.6 - graph-cut by Kolmogorov, Veksler et al
// New : MRF3.0
#include "graph.h"

// video cut
#include "video_cut_avec_contrainte_inclusion.h"

// Random generator
#include "randomness.h"

// Parameters
#include "parameter_class_medseq.h"

// Info about type limits
#include <climits>

// to save in PNG format
// #include "with_graphical_library.h"

namespace MRFSegm {



// A few simple tools

void error_display(char const * x){
  cout << endl << "Error message from graph cut : " << endl
       << x << endl;
}




// removed for MRF library compatibility issues (moved to a new one)

/* /\************************************************************************************************************\/ */
/* /\*************************** Minimization of binary MRF on a sequence of images  ****************************\/ */
/* /\************************************************************************************************************\/ */


/* // Minimize a given MRF  */
/* // Return the solution (one label per variable) */

/* // In Veksler, Zabih & Kolmogorov's code, the cost is of type float */
/* // Should modify mrf.h lines 45-46 to change cost type (supposing it works) */
/* vector< RecImage<bool> > minimize_MRF(const VideoEnergy<float>& E, bool verbose_base = true){ */



/* /\****************************************************************************************************\/ */
/* /\******************************** The same but with sequences of 3D images **************************\/ */
/* /\****************************************************************************************************\/ */

/* vector< Volume<bool> > minimize_MRF(const VideoVolumeEnergy<float>& E, bool verbose_base = true){ */



/****************************************************************************************************/
/************************** Coupled segmentation of sequences of 3D images **************************/
/****************************************************************************************************/

// List of types (for initial data and weights in the graph), to know the possible ranges of values available

// General structure
template <class T>  // T is the type of the data
struct FlowType {
  typedef T to_use; // type to use instead, if type T is too small
  typedef T total;  // Given data of type T, the flow computed is of type FlowType<T>::total.
  T valmax(){return 1;}   // and the probabilities or interaction values have maximal value FlowType<T>.valmax().  
  to_use valmax_to_use(){return 1;}  
  T zero(){return 0;} // Data values are represented in [zero, valmax] if positive, [2 zero - valmax, zero] if negative
};

// Useful instantiations
template<>
struct FlowType<unsigned char> {
  typedef short to_use;
  //typedef unsigned char to_use;
  typedef int  total;
  unsigned char valmax(){return UCHAR_MAX;}
  to_use valmax_to_use(){return SHRT_MAX;}  
  //  to_use valmax_to_use(){return UCHAR_MAX;}  
  unsigned char zero(){return UCHAR_MAX/2;}
 } ;

template<>
struct FlowType<short> {
  typedef short to_use;
  typedef int total;
  short valmax(){return SHRT_MAX;}
  to_use valmax_to_use(){return SHRT_MAX;}  
  short zero(){return SHRT_MAX/2;}
 } ;

template<>
struct FlowType<unsigned short> {
  typedef unsigned short to_use;
  typedef int total;
  unsigned short valmax(){return USHRT_MAX;}
  to_use valmax_to_use(){return USHRT_MAX;}  
  unsigned short zero(){return USHRT_MAX/2;}
 } ;

template<>
struct FlowType<float> {
  typedef float to_use;
  typedef float total;
  float valmax(){return 1;}
  to_use valmax_to_use(){return 1;}  
  float zero(){return 0;}
 } ;

 



// Given the energy and inclusion constraints between sequences, returns the optimal segmentation.
// See bla3M() as an example of use with a random energy.

// first vector< ... > : sequence index (MR, CT, ...)
// second vector< ... > : time index
// hence the result R reads R[sequence][time](x,y,z) 

// Inclusions is the list of sequence inclusion : (5,2) belongs to the list means foreground objet in sequence 5 is included in the foreground object of sequence 2

// empty_data : if true, will free E in order to gain space

// you should not need to read the code of this function

template <class datatype>
vector< vector< Volume<bool> > > minimize_MRF(vector< VideoVolumeEnergy<datatype> >& E, const list< pair<int,int> >& Inclusions, bool verbose_base = true, bool empty_data = false){

  /********* let's build the graph *******/
  
  /* // How it works in the binary graph cut library by Kolmogorov & al
     void Graph::add_edge(node_id from, node_id to, captype cap, captype rev_cap)
     void Graph::set_tweights(node_id i, captype cap_source, captype cap_sink)

     void* n1 = G.add_node();
     void* n2 = G.add_node();
     G.add_edge(n1, n2, 3, 3);
     G.set_tweights(n1, 7, 1);
     G.set_tweights(n2, 7, 1);
     
     float energy = G.maxflow();
     
     int a = G.what_segment(n1); // 0 = source, 1 = sink
  */
  
  int nb_seq = E.size(); // number of sequences
  int ns = E[0].neighborhoodsize/2; // neighborhood size has to be 26 ! // supposing all sequences consider same neigborhood
  
  //Building nodes and links to source/sink

  //  typedef Graph<typename datatype, typename datatype, typename FlowType<typename datatype>::total> GraphType;
  //typedef Graph<datatype, datatype, datatype> GraphType;
  
  typedef typename FlowType<datatype>::total flowtype;
  
  typedef Graph<datatype, datatype, flowtype> GraphType;
  
  //  typename FlowType<datatype>::total a;

  int nb_nodes_estim = (nb_seq * E[0].nt * E[0].nx * E[0].ny * E[0].nz + 2);
  GraphType *G = new GraphType(nb_nodes_estim, nb_nodes_estim * (ns + 4)); // estimated number of nodes and edges. +4 : 2 links (source,sink) + infinite in time + infinite towards another sequence
  
  
  // Storing addresses of nodes
  typedef typename GraphType::node_id Nnode; // that's a node in a graph (well, its address)
  vector< vector< Volume<Nnode> > > AdNod(nb_seq);
  for (int s = 0; s<nb_seq; s++){
    AdNod[s].resize(E[s].nt);
    for (int t = 0; t<E[s].nt; t++)
      AdNod[s][t].allocate(E[s].nx, E[s].ny, E[s].nz);
  }


  // Creating the graph, node by node

  for (int s = 0; s<nb_seq; s++)
    for (int t = 0; t<E[s].nt; t++){

    // creating nodes and adding potentials
      for (int i = 0; i<E[s].nx; i++)
	for (int j = 0; j<E[s].ny; j++)
	  for (int k = 0; k<E[s].nz; k++){
	    AdNod[s][t](i,j,k) = G->add_node();
    
	    datatype poids = E[s].Frames[t].Pot(i,j,k); // potential at point (i,j,k) in frame t
	    if (poids > 0)
	      G->add_tweights(AdNod[s][t](i,j,k), poids, 0); // NB : energy is multiplied by 2 in order not to be uselessly disturbed (????)
	    else
	      G->add_tweights(AdNod[s][t](i,j,k), 0, fabs(poids) );
	    //	cout << "poids : " << poids << endl;
	  }


      if (empty_data){
	//	for (int s = 0; s<nb_seq; s++)
	//	  for (int t = 0; t<E[s].nt; t++){
	    E[s].Frames[t].Pot.empty();
      }
      
    // interaction term within one image
      for (int i = 0; i<E[s].nx; i++)
	for (int j = 0; j<E[s].ny; j++)
	  for (int k = 0; k<E[s].nz; k++)

	    for (int d = 0; d<ns; d++){ // loop on neighbors. By (forced) luck, the 13 smallest indices within the 26 neighbors form precisely the good hald of the links to consider. 
	      int dx = (d%3) - 1 ; int dy = ((d%9)/3) - 1; int dz = (d/9) -1;  // supposing neighborhood size = 26 (Rubik's cube)

	      // who is my neighbor ?
	      int i0 = i+dx; int j0 = j+dy; int k0 = k+dz;
	      
	      if ( (i0 >= 0) && (j0 >= 0) && (k0 >= 0) && (i0 < E[s].nx) && (j0 < E[s].ny) && (k0 < E[s].nz) ) { // most of these tests are useless
		datatype poidsi = E[s].Frames[t].Inter(i,j,k)[d];
		G->add_edge( AdNod[s][t](i,j,k), AdNod[s][t](i0,j0,k0), poidsi, poidsi);
	      }
	    }

      if (empty_data){
	    E[s].Frames[t].Inter.empty();
	    E[s].Frames[t].nx = 0;
	    E[s].Frames[t].ny = 0;
	    E[s].Frames[t].nz = 0;
      }

      
    }
  
  



  // Constraint links to enforce shape shrinking only, within each sequence
  
  datatype infini = 20000; // attention à ce que l'énergie soit bien inférieure à cette valeur, garantie que tout se passe bien
  if (verbose_base)
    cout << "The infinite is " << infini << endl;

  for (int s = 0; s<nb_seq; s++)
    for (int t = 0; t<E[s].nt-1; t++)
      for (int i = 0; i<E[s].nx; i++)
	for (int j = 0; j<E[s].ny; j++)
	  for (int k = 0; k<E[s].nz; k++)
	    G->add_edge( AdNod[s][t](i,j,k), AdNod[s][t+1](i,j,k), 0, infini);

  //  cout << "Size of float : " << sizeof(float) << endl;
  //  cout << "Size of double : " << sizeof(double) << endl;




  // Constraint links to enforce shape inclusion between different sequences

  for (list<pair<int,int> >::const_iterator it = Inclusions.begin(); it != Inclusions.end(); it++){
    int inside = (*it).first;
    int containing = (*it).second;

    for (int t = 0; t<E[inside].nt; t++) // loop on all pixels
      for (int i = 0; i<E[inside].nx; i++)
	for (int j = 0; j<E[inside].ny; j++)
	  for (int k = 0; k<E[inside].nz; k++)
	    G->add_edge( AdNod[inside][t](i,j,k), AdNod[containing][t](i,j,k), 0, infini);

  }




  /****** We apply the graph cut algorithm ******/

  flowtype alors = G->maxflow(); 
  if (verbose_base)
    cout << "Energy at minimum (graph cut) : " << alors << endl;

  // Interprétation...
  
  vector< vector< Volume<bool> > > Resultat(nb_seq);

  for (int s = 0; s<nb_seq; s++){
    Resultat[s].resize(E[s].nt);
    for (int t = 0; t<E[s].nt; t++)
      Resultat[s][t].allocate(E[s].nx, E[s].ny, E[s].nz);
  
    for (int t = 0; t<E[s].nt; t++)
      for (int i = 0; i<E[s].nx; i++)
	for (int j = 0; j<E[s].ny; j++)
	  for (int k = 0; k<E[s].nz; k++)
	    Resultat[s][t](i,j,k) = ( G->what_segment(AdNod[s][t](i,j,k)) == GraphType::SOURCE ) ; // 0 = source, 1 = sink    
  }


  if ((verbose_base) && (! empty_data)){
    double energy_min = evaluate_energy(E, Resultat);
    cout << "Cost of this labelling choice : " << energy_min << endl;
  }
  
  // Cleaning

  for (int s = 0; s<nb_seq; s++){
    for (int t = 0; t<E[s].nt; t++)
      AdNod[s][t].empty();
    AdNod[s].empty();
  }
  AdNod.empty();
  
  
  /**** We display the solution (in text mode...) ****/

  //  cout << "Result : " << Resultat << endl;
  //  
  //  for (int i = 0; i<E.nb_var; i++)
  //    cout << "Pot " << i << " : " << E.Pot[i] << endl;
  //    
  //  for (int i = 0; i<E.nb_var; i++)
  //    for (int j = 0; j<int(E.Neighbors[i].size()); j++)
  //      if (i< E.Neighbors[i][j].first ) 
  //	cout << "M_(" << i << "," << E.Neighbors[i][j].first << ") : " << E.Interact[ E.Neighbors[i][j].second ].first << endl;


  return Resultat;
    
}







/**********************************************************************************************************************/
/*********************************** Reading & building graph at the same time ****************************************/
/**********************************************************************************************************************/

// easier to do with stringstream...
 string convertInt(int number)
 {
   if (number == 0)
     return "0";
   string temp="";
   string returnvalue="";
   while (number>0)
     {
       temp+=number%10+48;
       number/=10;
     }
   for (unsigned int i=0;i<temp.length();i++)
     returnvalue+=temp[temp.length()-i-1];
   return returnvalue;
 }



// Idem as previously, but with data read within the function, to spare memory space. 
// Also, 3D neighborhood is better built (coefficients), 
// and results are saved in .png files (if the graphical library is installed)

// first vector< ... > : sequence index (MR, CT, ...)
// second vector< ... > : time index
// hence the result R reads R[sequence][time](x,y,z) 

template <class probatype, class datatype>
  vector< vector< Volume<bool> > >  medical_scans(const Params& P){

  bool verbose_base = P.verbose_base;

  // ************** Data size and parameters ************** //

  int nb_seq = P.nb_seq; // number of sequences
  int ns = 26/2; // neighborhood size has to be 26 ! (divided by 2 for redundancy reasons)


  int nt = P.nt; // number of frames (should be read directly from data)

  int H = P.H; // size of original 3D images (height width depth) (should be read directly from data)
  int W = P.W;
  int D = P.D;
  int HW = H * W;

  double zoom_factor = P.zoom_factor; // to downsample the original data (to fit the computer's memory...)
  double zoom_inv = 1./zoom_factor;

  int nx = H * zoom_factor; // reduced size of down-sampled 3D images 
  int ny = W * zoom_factor;
  int nz = D * zoom_factor;
  int nxy = nx * ny;



  // ******** Inclusion constraints ********** //

  list<pair<int,int> > Inclusions = P.Inclusions; // list of (any number of) inclusions 
  // Inclusions is the list of sequence inclusion : (5,2) belongs to the list means foreground objet in sequence 5 is included in the foreground object of sequence 2
  //  Inclusions.push_back( pair<int,int>(0,1) ); // foreground in first sequence is included in the one of the second sequence (remember that 0 is the first sequence... and the last one is nb_seq-1).



  // ******** Setting neighborhood ********** //

  double coeff_distance[3]; // neighborhood coefficients in 3D
  coeff_distance[0] = 1;
  coeff_distance[1] = 1./sqrt(2.);
  coeff_distance[2] = 1./sqrt(3.);
  double coeff_tot = 6 * coeff_distance[0] + 12 * coeff_distance[1] + 8 * coeff_distance[2];



  // ********* Parameters for spatial interaction ********** //

  FlowType<datatype> FG; 
  double coeff_interactions = P.interaction_coeff / coeff_tot; // **** IMPORTANT PARAMETER **** : relative weight of interactions w.r.t. individual potentials (global factor)

  double sigma_inv = P.inverse_sigma_fact / FG.valmax(); // In the sequel, the interaction is defined using a Gaussian kernel on image intensity difference with the neighbors (see its use below): a standard deviation sigma thus needs to be defined. sigma_inv stands for its inverse. But you can define the interaction you want, you don't have to use a Gaussian kernel.


  // ******* Types of weights (double, int, short...) ********* //

  FlowType<probatype> F; // types used to build the graphe (short, int, float, etc.) and extreme values
  typedef typename FlowType<probatype>::total flowtype;
  typedef typename FlowType<probatype>::to_use edgetype;
  typedef Graph<edgetype, edgetype, flowtype> GraphType;
  probatype valmax = F.valmax();
  edgetype amplitude = F.valmax_to_use();
  probatype zero = F.zero();


  // ******* Temporal interactions ********* //

  // backward temporal interaction, to make segmentation results better at time extremes (doesn't change anything for other frames), as a shrinking constraint does not constrain the first frame at all... while this requires first & second frame to be quite similar. Can be set to 0 if none is wanted.
  edgetype lien_temporel =  0; // edgetype( 0.25 * double(valmax) );  



  // ***********  Summary  *********** //

  cout << "Data : " << nb_seq << " sequences of " << nt << " images, of initial size " << H << "x" << W << "x" << D << ", transformed into " << nx << "x" << ny << "x" << nz << " after a scaling factor " << zoom_factor << "." << endl;
  
  cout << "Parameters : sigma_inv = " << sigma_inv << " and coeff_interactions = " << coeff_interactions <<  "; temporal links : " << lien_temporel << "." << endl;
  for (list<pair<int,int> >::const_iterator it = Inclusions.begin(); it != Inclusions.end(); it++){
    int inside = (*it).first;
    int containing = (*it).second;
    cout << "Sequence " << inside+1 << " has foreground included in the one of sequence " << containing+1 << "." << endl;
  }
  cout << "Ready to go!" << endl;



  //Building nodes and links to source/sink

  // ****** Allocating the graph ******** // (possible memory issues)

  int nb_nodes_estim = (nb_seq * nt * nx * ny * nz + 2); // +2 : source/sink ?
  GraphType *G;
  try {  G = new GraphType(nb_nodes_estim, nb_nodes_estim * (ns + 4)); } // estimated number of nodes and edges
  catch (bad_alloc& bad){  cout << "Graph cut library failed allocating memory for " << nb_nodes_estim << " nodes and " <<  nb_nodes_estim * (ns + 4) << " vertices : " << bad.what() << endl; }
  cout << "Memory allocated for graph." << endl;

  // Storing addresses of nodes
  typedef typename GraphType::node_id Nnode; // that's a node in a graph (well, its address)
  vector< vector< Volume<Nnode> > > AdNod(nb_seq);
  for (int s = 0; s<nb_seq; s++){
    AdNod[s].resize(nt);
    for (int t = 0; t<nt; t++)
      AdNod[s][t].allocate(nx, ny, nz);
  }



  // ********* Creating the graph, node by node ************** //

  for (int s = 0; s<nb_seq; s++){

    // indicate here the correspondence between sequence number and actual numbering in data files to be read...
    //    int mod=1+2*s; // sequence 0 is modality 1; sequence 1 is modality 3

    probatype *IM = new probatype[H*W*D];   // space to store one 3D image

    for (int t = 0; t<nt; t++){

      // ******* load corresponding image (probability map) ********** (do it your own way, this is just an example)
      // remember that probatype is the type of the probabilities to be read/stored; it is defined when calling the function medical_scans.

      //      string NAME = "/user/gcharpia/home/Dropbox/Iceberg/TUMOR/OUT/probas/973_PR_M"+convertInt(mod)+"_T"+convertInt(t+1);
      //      string NAME = "/user/gcharpia/home/dataset/Icebergs_et_Tumeurs_Dropbox/TUMOR/OUT/973_PR_M"+convertInt(mod)+"_T"+convertInt(t+1);
      // WARNING : files on DropBox with these names are actually incorrect!!!...
      string NAME = P.input_prefix + P.input_prob_files[s] + convertInt(t+P.t0); // + P.input_suffix...
      FILE* f = fopen((char*)NAME.c_str(),"r");
      fread(IM,sizeof(probatype), H*W*D, f);
      fclose(f);
      probatype* pointeur = IM;

      edgetype zerob = 1. * zero; // needed to know the range of values given the type "probatype". Zerob is the 0, i.e. values v below zerob are interpreted as negative : - |v-zerob|, and values above are positive : |v-zerob|.

      // creating nodes and adding potentials
      for (int i = 0; i<nx; i++)
  	for (int j = 0; j<ny; j++)
  	  for (int k = 0; k<nz; k++){
  	    AdNod[s][t](i,j,k) = G->add_node();  // keep this
    
  	    // Individual potentials : put in poids the potential at point (i,j,k) in frame t (here read from image loaded)
  	    probatype poids = IM[ int(i*zoom_inv) + (int(j*zoom_inv))*H + (int(k*zoom_inv))*HW ];  // *pointeur; //E[s].Frames[t].Pot(i,j,k); // potential at point (i,j,k) in frame t

	    // keep the following
  	    if (poids > zerob)
  	      G->add_tweights(AdNod[s][t](i,j,k), edgetype(poids-zerob), 0);
  	    else
  	      G->add_tweights(AdNod[s][t](i,j,k), 0, edgetype(zerob-poids) );
  	    pointeur++;
  	  }
    }



    // ****** interaction term within one image *******

    delete[] IM; // types for probas and interactions might be different !
    datatype* IM2 = new datatype[H*W*D];   // space to store one 3D image



    for (int t = 0; t<nt; t++){

    // ******* load corresponding image (intensity image) ********* //

      //      string NAME = "/user/gcharpia/home/Dropbox/Iceberg/TUMOR/OUT/973_IM_M"+convertInt(mod)+"_T"+convertInt(t+1);
      //      string NAME = "/user/gcharpia/home/dataset/Icebergs_et_Tumeurs_Dropbox/TUMOR/OUT/973_IM_M"+convertInt(mod)+"_T"+convertInt(t+1);
      string NAME = P.input_prefix + P.input_inter_files[s] + convertInt(t+P.t0);
      FILE* f = fopen((char*)NAME.c_str(),"r");
      fread(IM2,sizeof(datatype), H*W*D, f);
      fclose(f);
      datatype* pointeur = IM2; // point to the first bin of the data


      for (int i = 0; i<nx; i++)
  	for (int j = 0; j<ny; j++)
  	  for (int k = 0; k<nz; k++){

	    // put here the 
  	    datatype im_value = IM2[ int(i*zoom_inv) + (int(j*zoom_inv))*H + (int(k*zoom_inv))*HW ]; //*pointeur;


	    // Loop on neighbors 
  	    for (int d = 0; d<ns; d++){ // By (forced) luck, the 13 smallest indices within the 26 neighbors form precisely the good half of the links to consider (remember that an interaction between two points has to be considered only once, whereas we iterate on all points)

  	      int dx = (d%3) - 1 ; int dy = ((d%9)/3) - 1; int dz = (d/9) -1;  // supposing neighborhood size = 26 (Rubik's cube)

  	      // who is my neighbor ?
  	      int i0 = i+dx; int j0 = j+dy; int k0 = k+dz;
	      
  	      if ( (i0 >= 0) && (j0 >= 0) && (k0 >= 0) && (i0 < nx) && (j0 < ny) && (k0 < nz) ) { // most of these tests are useless
  		// --> interaction term between 2 neighbors, namely (i,j,k) and (i+dx,j+dy,k+dz)=:(i0,j0,k0) : put the quantity you want (here a Gaussian function), weighted by a global factor coeff_interactions and by neighborhood geometry coeff_distance. That quantity has to fit the weight type and thus be between 0 and valmax-1.

  		edgetype poidsi = edgetype( coeff_interactions * coeff_distance[abs(dx) + abs(dy) + abs(dz) - 1] * exp(- square( sigma_inv * double(im_value - IM2[ int(i0*zoom_inv) + (int(j0*zoom_inv))*nx + (int(k0*zoom_inv))*nxy ]) )) * (double(valmax)) ); //E[s].Frames[t].Inter(i,j,k)[d];


		// Keep this as such.
  		G->add_edge( AdNod[s][t](i,j,k), AdNod[s][t](i0,j0,k0), poidsi, poidsi);
  	      }
  	    }

  	    pointeur++;
  	  }
    }
    delete[] IM2;
  }  





  // Constraint links to enforce shape shrinking only, within each sequence
  //                                 or GROWTH!!! 
 
  edgetype infini; // How much is infinite... in type edgetype.
  if ( (int(amplitude) > 20000) || (amplitude == 1))
    infini = edgetype(20000);
  else infini = amplitude; // make sure that the energy is smaller than this value, as a guarantee of infinite constraints to be infinite indeed.



  // ******** Keep the rest untouched ************ //

  if (verbose_base)
    cout << "Infinite is " << infini << endl;

  for (int s = 0; s<nb_seq; s++)
    for (int t = 0; t<nt-1; t++)
      for (int i = 0; i<nx; i++)
  	for (int j = 0; j<ny; j++)
  	  for (int k = 0; k<nz; k++)
  	    G->add_edge( AdNod[s][t](i,j,k), AdNod[s][t+1](i,j,k), infini, lien_temporel);



  // Constraint links to enforce shape inclusion between different sequences

  for (list<pair<int,int> >::const_iterator it = Inclusions.begin(); it != Inclusions.end(); it++){
    int inside = (*it).first;
    int containing = (*it).second;

    for (int t = 0; t<nt; t++) // loop on all pixels
      for (int i = 0; i<nx; i++)
  	for (int j = 0; j<ny; j++)
  	  for (int k = 0; k<nz; k++)
  	    G->add_edge( AdNod[inside][t](i,j,k), AdNod[containing][t](i,j,k), infini, 0);

  }



  cout << "Now, max flow..." << endl;
  /****** We apply the graph cut algorithm ******/

  flowtype alors = G->maxflow();
  if (verbose_base)
    cout << "Energy at minimum (graph cut) : " << alors << endl;



  // Interpretation...
  
  vector< vector< Volume<bool> > > Resultat(nb_seq);

  for (int s = 0; s<nb_seq; s++){
    Resultat[s].resize(nt);
    for (int t = 0; t<nt; t++)
      Resultat[s][t].allocate(nx, ny, nz);
  
    for (int t = 0; t<nt; t++)
      for (int i = 0; i<nx; i++)
  	for (int j = 0; j<ny; j++)
  	  for (int k = 0; k<nz; k++)
  	    Resultat[s][t](i,j,k) = ( G->what_segment(AdNod[s][t](i,j,k)) == GraphType::SOURCE ) ; // 0 = source, 1 = sink
  }

  cout << "Results allocated and filled." << endl;


  /* if ((verbose_base) && (! empty_data)){ */
  /*   double energy_min = evaluate_energy(E, Resultat); */
  /*   cout << "Cost of this labelling choice : " << energy_min << endl; */
  /* } */
  

  // Cleaning

  for (int s = 0; s<nb_seq; s++){
    for (int t = 0; t<nt; t++)
      AdNod[s][t].empty();
    AdNod[s].empty();
  }
  AdNod.empty();
  
  
  /**** We display the solution (in text mode...) ****/

  //  cout << "Result : " << Resultat << endl;
  //
  //  for (int i = 0; i<E.nb_var; i++)
  //    cout << "Pot " << i << " : " << E.Pot[i] << endl;
  //
  //  for (int i = 0; i<E.nb_var; i++)
  //    for (int j = 0; j<int(E.Neighbors[i].size()); j++)
  //      if (i< E.Neighbors[i][j].first )
  //	cout << "M_(" << i << "," << E.Neighbors[i][j].first << ") : " << E.Interact[ E.Neighbors[i][j].second ].first << endl;




  // Writing results slice by slice in png images or in binary format, depending on the graphical library begin installed or not

#ifdef I_want_to_use_the_graphical_library
  // Requires a graphical library... not installed by default (read the READ_ME to know more)

  int taille[3]; taille[0] = nx; taille[1] = ny; taille[2] = nz;
  char nomdir[3]; nomdir[0] = 'x'; nomdir[1] = 'y'; nomdir[2] = 'z';

  for (int s = 0; s<nb_seq; s++){
    for (int t = 0; t<nt; t++){
      for (int direction = 0; direction<3; direction++)
  	for (int i = 0; i<taille[direction]; i++){
  	  RecImage<bool> tranche = slice(Resultat[s][t], direction, i);
  	  char* ficout = new char[200];
	  //  	  sprintf(ficout, "%sseq%dt%d%ccut%d.png", nom.c_str(), s+1, t+1, nomdir[direction], i+1);
  	  sprintf(ficout, P.output_syntax_slice.c_str(), P.output_path.c_str(), s+1, t+P.t0, nomdir[direction], i+1);
  	  tranche.save(ficout);
  	  tranche.empty();
  	}
    }
  }
  
#else
  // No graphical library required

  if (P.save_volume){

    for (int s = 0; s<nb_seq; s++){
      for (int t = 0; t<nt; t++){
	
	char* ficout = new char[200];
	sprintf(ficout, P.output_syntax_volume.c_str(), P.output_path.c_str(), s+1, t+P.t0); // output_syntax is supposed to be of the form "%sseq%d_time%d.dat"
	sauve_binaire(Resultat[s][t], ficout); // save the 3D image in file "ficout", in binary format, as follows:
	// width  (size of integer)
	// height (size of integer)
	// depth  (size of integer)
	// foreach i, foreach j, foreach k, Matrix[i,j,k] (size of a boolean in the case here)
	// The content of the file will have to be known beforehand for reading : the fact that it is a matrix is not encoded in any header, same for the fact that elements are boolean.
      }
    }
  }
  

  // Slice by slice
  if (P.save_slice){
    int taille[3]; taille[0] = nx; taille[1] = ny; taille[2] = nz;
    char nomdir[3]; nomdir[0] = 'x'; nomdir[1] = 'y'; nomdir[2] = 'z';
    
    for (int s = 0; s<nb_seq; s++){
      for (int t = 0; t<nt; t++){
	for (int direction = 0; direction<3; direction++)
	  for (int i = 0; i<taille[direction]; i++){
	    RecImage<bool> tranche = slice(Resultat[s][t], direction, i);
	    char* ficout = new char[200];
	    //  	  sprintf(ficout, "%sseq%dt%d%ccut%d.png", nom.c_str(), s+1, t+1, nomdir[direction], i+1);
	    sprintf(ficout, P.output_syntax_slice.c_str(), P.output_path.c_str(), s+1, t+P.t0, nomdir[direction], i+1);
	    sauve_binaire(tranche, ficout); // save "tranche" in file "ficout", in binary format, as follows:
	    // width (size of integer)
	    // height (size of integer)
	    // foreach i, foreach j, Matrix[i,j] (size of a boolean in the case here)
	    // The content of the file will have to be known beforehand for reading : the fact that it is a matrix is not encoded in any header, same for the fact that elements are boolean.
	    tranche.empty();
	  }
      }
    }
  }

#endif
    

  // Returns the result...

  return Resultat;

  
}
 

/*******************************************************************************/
/********************   Shortcuts for instantiations   *************************/
/*******************************************************************************/
 
// Associated graphs have to be pre-declared in lib/MRF3.0/instances.inc

template <class ProbaType>
vector< vector< Volume<bool> > >  medical_scans(string data_type, const Params& P){

  if (data_type == "int"){   return medical_scans<ProbaType, int>(P);   }
  if (data_type == "short"){ return medical_scans<ProbaType, short>(P); }
  //  if (data_type == "char"){  return medical_scans<ProbaType, char>(P);  }
  // if (data_type == "unsigned_int"){    return medical_scans<ProbaType, unsigned int>(P);   }
  if (data_type == "unsigned_short"){  return medical_scans<ProbaType, unsigned short>(P); }
  if (data_type == "unsigned_char"){   return medical_scans<ProbaType, unsigned char>(P);  }
  if (data_type == "float"){  return medical_scans<ProbaType, float>(P); }
  if (data_type == "double"){ return medical_scans<ProbaType, double>(P);}


  cout << "Error : data type unknown : " << data_type << endl;
  return vector< vector< Volume<bool> > >();
}
 
vector< vector< Volume<bool> > >  medical_scans(string proba_type, string data_type, const Params& P){
  
  if (proba_type == "int"){    return medical_scans<int>(data_type, P); }
  if (proba_type == "short"){    return medical_scans<short>(data_type, P); }
  //  if (proba_type == "char"){    return medical_scans<char>(data_type, P); }
  //  if (proba_type == "unsigned_int"){    return medical_scans<unsigned int>(data_type, P); }
  if (proba_type == "unsigned_short"){    return medical_scans<unsigned short>(data_type, P); }
  if (proba_type == "unsigned_char"){    return medical_scans<unsigned char>(data_type, P); }
  if (proba_type == "float"){    return medical_scans<float>(data_type, P); }
  if (proba_type == "double"){    return medical_scans<double>(data_type, P); }

  cout << "Error : proba type unknown : " << proba_type << endl;
  return vector< vector< Volume<bool> > >();
}



} // End namespace

#endif
