// Global solutions to video cut with inclusion constraint in time

/***************** Library parameters ******************/
// Library parameters : comment if not suitable
// Use the graphical library, that needs to be installed first in that case ?
// Warning : every time you (un)comment this option, you have to update CMakeLists.txt accordingly and run cmake.
//#define I_want_to_use_the_graphical_library

// If your OS is Linux, you can have colors in terminals !
#define My_OS_is_GNU_Linux_and_I_like_colors


/***************** Including libraries *****************/
// Graphical library 
#ifdef I_want_to_use_the_graphical_library
 #include <Imagine/Graphics.h>
 #include <Imagine/Images.h>
 using namespace Imagine;
#endif

// If OS is linux, include first
// otherwise inlude second
#ifdef My_OS_is_GNU_Linux_and_I_like_colors
 #include "colors_in_terminal.h"
#else
 #include "no_colors_in_terminal.h"
#endif


// Personal library
#include "images.h"
#include "video_cut_avec_contrainte_inclusion.h"
#include "volumes.h"

// With graphical library ?
#ifdef I_want_to_use_the_graphical_library
 #include "with_graphical_library.h"
 // should have a look at this one even if not using this graphical library :
 #include "requires_graphical_library.h"  
#endif


// Minimization tools
#include "minimization.h"
#include "parameter_class_medseq.h"

// Others
#include <ctime>
//#include <time.h>
//#include <iomanip>









// Random test with multiple sequences of 3D volumes
// First build a random energy, then call the minimizer.
// *** See minimize_MRF for notations ***

void bla3M(){

  // Definition of an energy for 2 sequences of 5 frames of size 40x40x40.
  
  int nseq = 2;
  int nt = 5;
  int nx = 40;
  int ny = 40;
  int nz = 40;
  
  int ns; // neighborhood size (no choice, currently


  typedef short datatype;
  //typedef float datatype;
  //  typedef double datatype;

  vector< MRFSegm::VideoVolumeEnergy<datatype> > VSeq(nseq);

  cout << nseq << " sequences of images" << endl;

  for (int s = 0; s<nseq; s++){

    VSeq[s].allocate(nt,nx,ny,nz);
    
    cout << "Video (volume) number " << s+1 << " : " << nt << " frame" <<  ((nt<2)?"":"s") << ", de taille " << nx << " x " << ny << " x " << nz << endl;
    cout << "Size of neighborhood : " << VSeq[s].neighborhoodsize << endl;
    
    ns = VSeq[s].neighborhoodsize/2; // so, 13


    // Building a random energy
    
    for (int t = 0; t<nt; t++)
      for (int i = 0; i<nx; i++)
	for (int j = 0; j<ny; j++)
	  for (int k = 0; k<nz; k++){
	    
	   // Potential at (x,y,z) at time t : fill VSeq[s].Frames[t].Pot(i,j,k) 
	    if (i > nx/6 + t + s)
	      //	 if ( ( (i>=nx/3)&&(i<=2*nx/3) ) && ( (j>=ny/3)&&(j<=2*ny/3) ) )
	      VSeq[s].Frames[t].Pot(i,j,k) = 10;
	    else
	      VSeq[s].Frames[t].Pot(i,j,k) = -10;
	    
	    if ((t > 0)&&(t < nt-1))
	      VSeq[s].Frames[t].Pot(i,j,k) = 2*(rand()%3 - 1);
	   
	    
	    // Loop on neighbors
	    for (int d = 0; d<ns; d++){ // the first 13 neighbors with the enforced numbering of all 26 voisins are the good half of links to consider : loop on directions d
	      int dx = (d%3) - 1 ; int dy = ((d%9)/3) - 1; int dz = (d/9) -1;  // neighborhood size = 26 (Rubik's cube)
	      int i0 = i+dx; int j0 = j+dy; int k0 = k+dz;
	      
	      // just any stupid energy as an example
	      if ( (i0 >= 0) && (j0 >= 0) && (k0 >= 0) && (i0 < nx) && (j0 < ny) && (k0 < nz) ) { // la plupart de ces tests sont inutiles, c'est pour le folklore
	       
		// neighboring interaction, between (i,j,k) and (i,j,k)+direction d = (i0,j0,k0) in the same image, at time t
		// Fill VSeq[s].Frames[t].Inter(i,j,k)[d] for the first 13th directions
		// and multiply it by a relative weight of interactions vs. potentials, and by a neighborhood weight (that says that this pair of neighbors is closer than this other pair) if desired.

		// random example to define an energy : 
		if ( ( i == nx/2 - 2*t) ) // this is just a basic example... doesn't make sense.
		  VSeq[s].Frames[t].Inter(i,j,k)[d] = 1; 
		else 
		  VSeq[s].Frames[t].Inter(i,j,k)[d] = 1; //0.025;
	       
		if ((t > 0)&&(t < nt-1))
		  //	   if (t == 1)
		  VSeq[s].Frames[t].Inter(i,j,k)[d] = 1;
		
	      }
	    } // end of the loop on neighbors
	    
	    
	  } // end of the loop on pixels
  }


  // ********* define inclusion constraints **********
  list< pair<int,int> > List_inclusions;
  for (int s = 0; s<nseq-1; s++)
    List_inclusions.push_back( pair<int,int>(s,s+1) );



  // ********* Minimization ******** //

  cout << "Before minim" << endl;
  vector< vector< MRFSegm::Volume<bool> > > Resultat = MRFSegm::minimize_MRF(VSeq, List_inclusions, true, true);
  cout << "After minim" << endl;

   
  // ********* Result ***********

  // for (int s = 0; s<nseq; s++){
  //   cout << "Sequence number " << s+1 << " :" << endl;
  //   for (int t = 0; t<nt;t++){
  //     cout << "Image " << t << " :" << endl;
  //     //      cout << VSeq[s].Frames[t].Pot << endl; // initial potential
  //     //      cout << Resultat[s][t] << endl; // result
  //     cout << endl;
  //   }
  // }
  
  
   //     cout << Resultat[0] << endl;

}









/*************************************  The main function ***************************/


int main(int argc, char **argv){


  // To perform a random test without data :
  cout << "-------- Random test --------" << endl;

  /******** Randomness needed *********/  
  MRFSegm::randomness_initialization();

  // the test (see above) :
  bla3M();




  // Practical framework to perform real tests with data, with parameters defined in an external file :
  cout << "-------- Study case of 3D medical scans --------" << endl;
  // Note : the data related to the particular study described in params.dat is not provided, so this should segfault.


  /********* Reading parameters **********/
  // by default, parameter file is "params.dat"
  // otherwise, it is the first (and unique) argument on the command line
  MRFSegm::Params P;
  P.charge_params(argc, argv);

  // if willing to copy the parameter file in the result directory (to remember parameter used for that experiment)
  MRFSegm::sauver_params(P.nom_fichier_param, P.output_path);

  

  cout << "I compile and run !!!" << endl;


  // The functions to look at :
  MRFSegm::medical_scans(P.proba_type, P.data_type, P);

  //  MRFSegm::medical_scans<unsigned char, unsigned short>();
  // or
  // bla3M();

  cout << "End of the story." << endl;
  return 0;
  


/********* If using the graphical library **********/

// #ifdef I_want_to_use_the_graphical_library
//   cout << endl << endl << MRFSegm::BoldRed << "  ***************  " << MRFSegm::BoldGreen << "Using the Imagine++ graphical library!" << MRFSegm::BoldRed << "  ***************" << MRFSegm::TextNormal << endl << endl;

// ...

//   endGraphics(); // should be added at the end of the code, if and only if a window has been opened before

//   cout << endl << endl << "This is the end (again)..." << endl << endl;
// #endif


  return 0;
}


